"""Implements class that holds custom load and invoke function of a model"""

from __future__ import absolute_import
import abc


class InferenceSpec(abc.ABC):
    """Abstract base class for holding custom ``load``, ``invoke`` and ``prepare`` functions.

    Provides a skeleton for customization to override the methods
    ``load``, ``invoke`` and ``prepare``.
    """

    @abc.abstractmethod
    def load(self, model_dir: str):
        """Loads the model stored in model_dir and return the model object.

        Args:
            model_dir (str): Path to the directory where the model is stored.
        """

    @abc.abstractmethod
    def invoke(self, input_object: object, model: object):
        """Given model object and input, make inference and return the result.

        Args:
            input_object (object): The input to model
            model (object): The model object
        """

    def preprocess(self, input_data: object):
        """Custom pre-processing function"""

    def postprocess(self, predictions: object):
        """Custom post-processing function"""

    def prepare(self, *args, **kwargs):
        """Custom prepare function"""

    def get_model(self):
        """Return HuggingFace model name for inference spec"""



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Holds the util functions used for lineage tracking"""
from __future__ import absolute_import

import os
import time
import re
import logging
from typing import List, Optional, Union

from botocore.exceptions import ClientError

from sagemaker.core.helper.session_helper import Session
from sagemaker.core.shapes.shapes import ArtifactSummary
from sagemaker.core.resources import Artifact, Association
from sagemaker.serve.model_format.mlflow.constants import (
    MLFLOW_RUN_ID_REGEX,
    MODEL_PACKAGE_ARN_REGEX,
    S3_PATH_REGEX,
    MLFLOW_REGISTRY_PATH_REGEX,
)
from sagemaker.serve.utils.lineage_constants import (
    LINEAGE_POLLER_MAX_TIMEOUT_SECS,
    LINEAGE_POLLER_INTERVAL_SECS,
    TRACKING_SERVER_ARN_REGEX,
    TRACKING_SERVER_CREATION_TIME_FORMAT,
    MLFLOW_S3_PATH,
    MODEL_BUILDER_MLFLOW_MODEL_PATH_LINEAGE_ARTIFACT_TYPE,
    MLFLOW_LOCAL_PATH,
    MLFLOW_MODEL_PACKAGE_PATH,
    MLFLOW_RUN_ID,
    MLFLOW_REGISTRY_PATH,
    CONTRIBUTED_TO,
    ERROR,
    CODE,
    VALIDATION_EXCEPTION,
)

logger = logging.getLogger(__name__)


def _load_artifact_by_source_uri(
    source_uri: str,
    sagemaker_session: Session,
    source_types_to_match: Optional[List[str]] = None,
    artifact_type: Optional[str] = None,
) -> Optional[ArtifactSummary]:
    """Load lineage artifact by source uri

    Arguments:
        source_uri (str): The s3 uri used for uploading transfomred model artifacts.
        sagemaker_session (Session): Session object which manages interactions
            with Amazon SageMaker APIs and any other AWS services needed. If not specified, the
            function creates one using the default AWS configuration chain.
        source_types_to_match (Optional[List[str]]): A list of source type values to match against
            the artifact's source types. If provided, the artifact's source types must match this
            list.
        artifact_type (Optional[str]): The type of the lineage artifact.

    Returns:
        ArtifactSummary: The Artifact Summary for the provided S3 URI.
    """
    artifacts = Artifact.list(source_uri=source_uri, sagemaker_session=sagemaker_session)
    for artifact_summary in artifacts:
        if artifact_type is None or artifact_summary.artifact_type == artifact_type:
            if source_types_to_match:
                if artifact_summary.source.source_types is not None:
                    artifact_source_types = [
                        source_type["Value"] for source_type in artifact_summary.source.source_types
                    ]
                    if set(artifact_source_types) == set(source_types_to_match):
                        return artifact_summary
                else:
                    return None
            else:
                return artifact_summary

    return None


def _poll_lineage_artifact(
    s3_uri: str, artifact_type: str, sagemaker_session: Session
) -> Optional[ArtifactSummary]:
    """Polls lineage artifacts by s3 path.

    Arguments:
        s3_uri (str): The S3 URI to check for artifacts.
        artifact_type (str): The type of the lineage artifact.
        sagemaker_session (Session): Session object which manages interactions
            with Amazon SageMaker APIs and any other AWS services needed. If not specified, the
            function creates one using the default AWS configuration chain.

    Returns:
        Optional[ArtifactSummary]: The artifact summary if found, otherwise None.
    """
    logger.info("Polling lineage artifact for model data in %s", s3_uri)
    start_time = time.time()
    while time.time() - start_time < LINEAGE_POLLER_MAX_TIMEOUT_SECS:
        result = _load_artifact_by_source_uri(
            s3_uri, sagemaker_session, artifact_type=artifact_type
        )
        if result is not None:
            return result
        time.sleep(LINEAGE_POLLER_INTERVAL_SECS)


def _get_mlflow_model_path_type(mlflow_model_path: str) -> str:
    """Identify mlflow model path type.

    Args:
        mlflow_model_path (str): The string to be identified.

    Returns:
        str: Description of what the input string is identified as.
    """
    mlflow_run_id_pattern = MLFLOW_RUN_ID_REGEX
    mlflow_registry_id_pattern = MLFLOW_REGISTRY_PATH_REGEX
    sagemaker_arn_pattern = MODEL_PACKAGE_ARN_REGEX
    s3_pattern = S3_PATH_REGEX

    if re.match(mlflow_run_id_pattern, mlflow_model_path):
        return MLFLOW_RUN_ID
    if re.match(mlflow_registry_id_pattern, mlflow_model_path):
        return MLFLOW_REGISTRY_PATH
    if re.match(sagemaker_arn_pattern, mlflow_model_path):
        return MLFLOW_MODEL_PACKAGE_PATH
    if re.match(s3_pattern, mlflow_model_path):
        return MLFLOW_S3_PATH
    if os.path.exists(mlflow_model_path):
        return MLFLOW_LOCAL_PATH

    raise ValueError(f"Invalid MLflow model path: {mlflow_model_path}")


def _create_mlflow_model_path_lineage_artifact(
    mlflow_model_path: str,
    sagemaker_session: Session,
    source_types_to_match: Optional[List[str]] = None,
) -> Optional[Artifact]:
    """Creates a lineage artifact for the given MLflow model path.

    Args:
        mlflow_model_path (str): The path to the MLflow model.
        sagemaker_session (Session): The SageMaker session object.
        source_types_to_match (Optional[List[str]]): Artifact source types.

    Returns:
        Optional[Artifact]: The created lineage artifact, or None if an error occurred.
    """
    _artifact_name = _get_mlflow_model_path_type(mlflow_model_path)
    properties = dict(
        model_builder_input_model_data_type=_artifact_name,
    )
    try:
        source_types = [dict(SourceIdType="Custom", Value="ModelBuilderInputModelData")]
        if source_types_to_match:
            source_types += [
                dict(SourceIdType="Custom", Value=source_type)
                for source_type in source_types_to_match
                if source_type != "ModelBuilderInputModelData"
            ]

        return Artifact.create(
            source_uri=mlflow_model_path,
            source_types=source_types,
            artifact_type=MODEL_BUILDER_MLFLOW_MODEL_PATH_LINEAGE_ARTIFACT_TYPE,
            artifact_name=_artifact_name,
            properties=properties,
            sagemaker_session=sagemaker_session,
        )
    except ClientError as e:
        if e.response[ERROR][CODE] == VALIDATION_EXCEPTION:
            logger.info("Artifact already exists")
        else:
            logger.warning("Failed to create mlflow model path lineage artifact: %s", e)
            raise e


def _retrieve_and_create_if_not_exist_mlflow_model_path_lineage_artifact(
    mlflow_model_path: str,
    sagemaker_session: Session,
    tracking_server_arn: Optional[str] = None,
) -> Optional[Union[Artifact, ArtifactSummary]]:
    """Retrieves an existing artifact for the given MLflow model path or

    creates a new one if it doesn't exist.

    Args:
        mlflow_model_path (str): The path to the MLflow model.
        sagemaker_session (Session): Session object which manages interactions
            with Amazon SageMaker APIs and any other AWS services needed. If not specified, the
            function creates one using the default AWS configuration chain.
        tracking_server_arn (Optional[str]): The MLflow tracking server ARN.

    Returns:
        Optional[Union[Artifact, ArtifactSummary]]: The existing or newly created artifact,
            or None if an error occurred.
    """
    source_types_to_match = ["ModelBuilderInputModelData"]
    input_type = _get_mlflow_model_path_type(mlflow_model_path)
    if tracking_server_arn and input_type in [MLFLOW_RUN_ID, MLFLOW_REGISTRY_PATH]:
        match = re.match(TRACKING_SERVER_ARN_REGEX, tracking_server_arn)
        mlflow_tracking_server_name = match.group(4)
        describe_result = sagemaker_session.sagemaker_client.describe_mlflow_tracking_server(
            TrackingServerName=mlflow_tracking_server_name
        )
        tracking_server_creation_time = describe_result["CreationTime"].strftime(
            TRACKING_SERVER_CREATION_TIME_FORMAT
        )
        source_types_to_match += [tracking_server_arn, tracking_server_creation_time]
    _loaded_artifact = _load_artifact_by_source_uri(
        mlflow_model_path,
        sagemaker_session,
        source_types_to_match,
    )
    if _loaded_artifact is not None:
        return _loaded_artifact
    return _create_mlflow_model_path_lineage_artifact(
        mlflow_model_path,
        sagemaker_session,
        source_types_to_match,
    )


def _add_association_between_artifacts(
    mlflow_model_path_artifact_arn: str,
    autogenerated_model_data_artifact_arn: str,
    sagemaker_session: Session,
) -> None:
    """Add association between mlflow model path artifact and autogenerated model data artifact.

    Arguments:
        mlflow_model_path_artifact_arn (str): The mlflow model path artifact.
        autogenerated_model_data_artifact_arn (str): The autogenerated model data artifact.
        sagemaker_session (Session): Session object which manages interactions
            with Amazon SageMaker APIs and any other AWS services needed. If not specified, the
            function creates one using the default AWS configuration chain.
    """
    _association_type = CONTRIBUTED_TO
    _source_arn = mlflow_model_path_artifact_arn
    _destination_arn = autogenerated_model_data_artifact_arn
    try:
        logger.info(
            "Adding association with source_arn: "
            "%s, destination_arn: %s and association_type: %s.",
            _source_arn,
            _destination_arn,
            _association_type,
        )
        Association.create(
            source_arn=_source_arn,
            destination_arn=_destination_arn,
            association_type=_association_type,
            sagemaker_session=sagemaker_session,
        )
    except ClientError as e:
        if e.response[ERROR][CODE] == VALIDATION_EXCEPTION:
            logger.info("Association already exists")
        else:
            raise e


def _maintain_lineage_tracking_for_mlflow_model(
    mlflow_model_path: str,
    s3_upload_path: str,
    sagemaker_session: Session,
    tracking_server_arn: Optional[str] = None,
) -> None:
    """Maintains lineage tracking for an MLflow model by creating or retrieving artifacts.

    Args:
        mlflow_model_path (str): The path to the MLflow model.
        s3_upload_path (str): The S3 path where the transformed model data is uploaded.
        sagemaker_session (Session): Session object which manages interactions
            with Amazon SageMaker APIs and any other AWS services needed. If not specified, the
            function creates one using the default AWS configuration chain.
        tracking_server_arn (Optional[str]): The MLflow tracking server ARN.
    """
    artifact_for_transformed_model_data = _poll_lineage_artifact(
        s3_uri=s3_upload_path,
        artifact_type="ModelData",
        sagemaker_session=sagemaker_session,
    )
    if artifact_for_transformed_model_data:
        mlflow_model_artifact = (
            _retrieve_and_create_if_not_exist_mlflow_model_path_lineage_artifact(
                mlflow_model_path=mlflow_model_path,
                sagemaker_session=sagemaker_session,
                tracking_server_arn=tracking_server_arn,
            )
        )
        if mlflow_model_artifact:
            _mlflow_model_artifact_arn = (
                mlflow_model_artifact.artifact_arn
            )  # pylint: disable=E1101, disable=C0301
            _artifact_for_transformed_model_data_arn = (
                artifact_for_transformed_model_data.artifact_arn
            )  # pylint: disable=C0301
            _add_association_between_artifacts(
                mlflow_model_path_artifact_arn=_mlflow_model_artifact_arn,
                autogenerated_model_data_artifact_arn=_artifact_for_transformed_model_data_arn,
                sagemaker_session=sagemaker_session,
            )
        else:
            logger.warning(
                "Unable to add association between autogenerated lineage "
                "artifact for transformed model data and mlflow model path"
                " lineage artifacts."
            )
    else:
        logger.warning(
            "Lineage artifact for transformed model data is not auto-created within "
            "%s seconds, skipping creation of lineage artifact for mlflow model path",
            LINEAGE_POLLER_MAX_TIMEOUT_SECS,
        )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Holds constants used for interpreting MLflow models."""
from __future__ import absolute_import

DEFAULT_FW_USED_FOR_DEFAULT_IMAGE = "pytorch"
DEFAULT_PYTORCH_VERSION = {
    "py38": "1.12.1",
    "py39": "1.13.1",
    "py310": "2.2.0",
    "py311": "2.3.0",
}
MODEL_PACKAGE_ARN_REGEX = (
    r"^arn:aws:sagemaker:[a-z0-9\-]+:[0-9]{12}:model-package\/(.*?)(?:/(\d+))?$"
)
MLFLOW_RUN_ID_REGEX = r"^runs:/[a-zA-Z0-9]+(/[a-zA-Z0-9\-_\.]*)+$"
MLFLOW_REGISTRY_PATH_REGEX = r"^models:/[a-zA-Z0-9\-_\.]+[@/]?[a-zA-Z0-9\-_\.][/a-zA-Z0-9\-_\.]*$"
S3_PATH_REGEX = r"^s3:\/\/[a-zA-Z0-9\-_\.]+(?:\/[a-zA-Z0-9\-_\/\.]*)?$"
MLFLOW_TRACKING_ARN = "MLFLOW_TRACKING_ARN"
MLFLOW_MODEL_PATH = "MLFLOW_MODEL_PATH"
MLFLOW_METADATA_FILE = "MLmodel"
MLFLOW_PIP_DEPENDENCY_FILE = "requirements.txt"
MLFLOW_PYFUNC = "python_function"
MLFLOW_FLAVOR_TO_PYTHON_PACKAGE_MAP = {
    "sklearn": "scikit-learn",
    "pytorch": "torch",
    "tensorflow": "tensorflow",
    "keras": "tensorflow",
    "xgboost": "xgboost",
    "lightgbm": "lightgbm",
    "h2o": "h2o",
    "spark": "pyspark",
    "onnx": "onnxruntime",
}
TENSORFLOW_SAVED_MODEL_NAME = "saved_model.pb"
FLAVORS_WITH_FRAMEWORK_SPECIFIC_DLC_SUPPORT = {
    "sklearn": "sklearn",
    "pytorch": "pytorch",
    "xgboost": "xgboost",
    "tensorflow": "tensorflow",
    "keras": "tensorflow",
    "spark": "sparkml"
}
FLAVORS_DEFAULT_WITH_TF_SERVING = ["keras", "tensorflow"]



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""This module is for SageMaker inference.py."""

from __future__ import absolute_import
import asyncio
import os
import platform
import cloudpickle
import logging
from pathlib import Path
from sagemaker.serve.validations.check_integrity import perform_integrity_check


logger = LOGGER = logging.getLogger("sagemaker")


def initialize_custom_orchestrator():
    """Initializes the custom orchestrator."""
    code_dir = os.getenv("SAGEMAKER_INFERENCE_CODE_DIRECTORY", None)
    serve_path = Path(code_dir).joinpath("serve.pkl")
    with open(str(serve_path), mode="rb") as pkl_file:
        return cloudpickle.load(pkl_file)


def _run_preflight_diagnostics():
    _py_vs_parity_check()
    _pickle_file_integrity_check()


def _py_vs_parity_check():
    container_py_vs = platform.python_version()
    local_py_vs = os.getenv("LOCAL_PYTHON")

    if not local_py_vs or container_py_vs.split(".")[1] != local_py_vs.split(".")[1]:
        logger.warning(
            f"The local python version {local_py_vs} differs from the python version "
            f"{container_py_vs} on the container. Please align the two to avoid unexpected behavior"
        )


def _pickle_file_integrity_check():
    with open("/opt/ml/model/code/serve.pkl", "rb") as f:
        buffer = f.read()

    metadata_path = Path("/opt/ml/model/code/metadata.json")
    perform_integrity_check(buffer=buffer, metadata_path=metadata_path)


_run_preflight_diagnostics()
custom_orchestrator, _ = initialize_custom_orchestrator()


async def handler(request):
    """Custom service entry point function.

    :param request: raw input from request
    :return: outputs to be send back to client
    """
    if asyncio.iscoroutinefunction(custom_orchestrator.handle):
        return await custom_orchestrator.handle(request.body)
    else:
        return custom_orchestrator.handle(request.body)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Shared resources for prepare step of model deployment"""

from __future__ import absolute_import
import logging

from sagemaker.serve.model_server.tgi.prepare import _copy_jumpstart_artifacts
from sagemaker.serve.utils.local_hardware import _check_disk_space, _check_docker_disk_usage

from pathlib import Path
import shutil
from typing import List

from sagemaker.core.helper.session_helper import Session
from sagemaker.serve.spec.inference_spec import InferenceSpec
from sagemaker.serve.detector.dependency_manager import capture_dependencies
from sagemaker.serve.validations.check_integrity import (
    generate_secret_key,
    compute_hash,
)
from sagemaker.core.remote_function.core.serialization import _MetaData

logger = logging.getLogger(__name__)


def _create_dir_structure(model_path: str) -> tuple:
    """Create the expected model directory structure for the Multi Model server"""
    model_path = Path(model_path)
    if not model_path.exists():
        model_path.mkdir(parents=True)
    elif not model_path.is_dir():
        raise ValueError("model_dir is not a valid directory")

    code_dir = model_path.joinpath("code")
    code_dir.mkdir(exist_ok=True, parents=True)

    _check_disk_space(model_path)
    _check_docker_disk_usage()

    return model_path, code_dir


def prepare_mms_js_resources(
    model_path: str,
    js_id: str,
    shared_libs: List[str] = None,
    dependencies: str = None,
    model_data: str = None,
) -> tuple:
    """Prepare serving when a JumpStart model id is given

    Args:
        model_path (str) : Argument
        js_id (str): Argument
        shared_libs (List[]) : Argument
        dependencies (str) : Argument
        model_data (str) : Argument

    Returns:
        ( str ) :

    """
    model_path, code_dir = _create_dir_structure(model_path)

    return _copy_jumpstart_artifacts(model_data, js_id, code_dir)


def prepare_for_mms(
    model_path: str,
    shared_libs: List[str],
    dependencies: dict,
    session: Session,
    image_uri: str,
    inference_spec: InferenceSpec = None,
) -> str:
    """Prepares for InferenceSpec using model_path, writes inference.py, and captures dependencies to generate secret_key.

    Args:to
        model_path (str) : Argument
        shared_libs (List[]) : Argument
        dependencies (dict) : Argument
        session (Session) : Argument
        inference_spec (InferenceSpec, optional) : Argument
            (default is None)
    Returns:
        ( str ) : secret_key
    """
    model_path = Path(model_path)
    if not model_path.exists():
        model_path.mkdir()
    elif not model_path.is_dir():
        raise Exception("model_dir is not a valid directory")

    if inference_spec:
        inference_spec.prepare(str(model_path))

    code_dir = model_path.joinpath("code")
    code_dir.mkdir(exist_ok=True)

    shutil.copy2(Path(__file__).parent.joinpath("inference.py"), code_dir)

    logger.info("Finished writing inference.py to code directory")

    shared_libs_dir = model_path.joinpath("shared_libs")
    shared_libs_dir.mkdir(exist_ok=True)
    for shared_lib in shared_libs:
        shutil.copy2(Path(shared_lib), shared_libs_dir)

    capture_dependencies(dependencies=dependencies, work_dir=code_dir)

    secret_key = generate_secret_key()
    with open(str(code_dir.joinpath("serve.pkl")), "rb") as f:
        buffer = f.read()
    hash_value = compute_hash(buffer=buffer, secret_key=secret_key)
    with open(str(code_dir.joinpath("metadata.json")), "wb") as metadata:
        metadata.write(_MetaData(hash_value).to_json())

    return secret_key



"""This module is for SageMaker inference.py."""

from __future__ import absolute_import
import os
import io
import json
import cloudpickle
import shutil
import platform
from pathlib import Path
from sagemaker.serve.validations.check_integrity import perform_integrity_check
import logging

logger = logging.getLogger(__name__)

schema_builder = None
SHARED_LIBS_DIR = Path(__file__).parent.parent.joinpath("shared_libs")
SERVE_PATH = Path(__file__).parent.joinpath("serve.pkl")
METADATA_PATH = Path(__file__).parent.joinpath("metadata.json")


def input_handler(data, context):
    """Pre-process request input before it is sent to TensorFlow Serving REST API

    Args:
        data (obj): the request data, in format of dict or string
        context (Context): an object containing request and configuration details
    Returns:
        (dict): a JSON-serializable dict that contains request body and headers
    """
    read_data = data.read()
    deserialized_data = None
    try:
        if hasattr(schema_builder, "custom_input_translator"):
            deserialized_data = schema_builder.custom_input_translator.deserialize(
                io.BytesIO(read_data), context.request_content_type
            )
        else:
            deserialized_data = schema_builder.input_deserializer.deserialize(
                io.BytesIO(read_data), context.request_content_type
            )

    except Exception as e:
        logger.error("Encountered error: %s in deserialize_request." % e)
        raise Exception("Encountered error in deserialize_request.") from e

    try:
        return json.dumps({"instances": _convert_for_serialization(deserialized_data)})
    except Exception as e:
        logger.error(
            "Encountered error: %s in deserialize_request. "
            "Deserialized data is not json serializable. " % e
        )
        raise Exception("Encountered error in deserialize_request.") from e


def output_handler(data, context):
    """Post-process TensorFlow Serving output before it is returned to the client.

    Args:
        data (obj): the TensorFlow serving response
        context (Context): an object containing request and configuration details
    Returns:
        (bytes, string): data to return to client, response content type
    """
    if data.status_code != 200:
        raise ValueError(data.content.decode("utf-8"))

    response_content_type = context.accept_header
    prediction = data.content
    try:
        prediction_dict = json.loads(prediction.decode("utf-8"))
        if hasattr(schema_builder, "custom_output_translator"):
            return (
                schema_builder.custom_output_translator.serialize(
                    prediction_dict["predictions"], response_content_type
                ),
                response_content_type,
            )
        else:
            return schema_builder.output_serializer.serialize(prediction_dict["predictions"]), response_content_type
    except Exception as e:
        logger.error("Encountered error: %s in serialize_response." % e)
        raise Exception("Encountered error in serialize_response.") from e


def _run_preflight_diagnostics():
    _py_vs_parity_check()
    _pickle_file_integrity_check()


def _py_vs_parity_check():
    container_py_vs = platform.python_version()
    local_py_vs = os.getenv("LOCAL_PYTHON")

    if not local_py_vs or container_py_vs.split(".")[1] != local_py_vs.split(".")[1]:
        logger.warning(
            f"The local python version {local_py_vs} differs from the python version "
            f"{container_py_vs} on the container. Please align the two to avoid unexpected behavior"
        )


def _pickle_file_integrity_check():
    with open(SERVE_PATH, "rb") as f:
        buffer = f.read()

    perform_integrity_check(buffer=buffer, metadata_path=METADATA_PATH)


def _set_up_schema_builder():
    """Sets up the schema_builder object."""
    global schema_builder
    with open(SERVE_PATH, "rb") as f:
        schema_builder = cloudpickle.load(f)


def _set_up_shared_libs():
    """Sets up the shared libs path."""
    if SHARED_LIBS_DIR.exists():
        # before importing, place dynamic linked libraries in shared lib path
        shutil.copytree(SHARED_LIBS_DIR, "/lib", dirs_exist_ok=True)


def _convert_for_serialization(deserialized_data):
    """Attempt to convert non-serializable objects to a serializable form.

    Args:
        deserialized_data: The object to convert.

    Returns:
        The converted object if it was not originally serializable, otherwise the original object.
    """
    import numpy as np
    import pandas as pd

    if isinstance(deserialized_data, np.ndarray):
        return deserialized_data.tolist()
    elif isinstance(deserialized_data, pd.DataFrame):
        return deserialized_data.to_dict(orient="list")
    elif isinstance(deserialized_data, pd.Series):
        return deserialized_data.tolist()
    return deserialized_data


# on import, execute
_run_preflight_diagnostics()
_set_up_schema_builder()
_set_up_shared_libs()
